from typing import Any, Dict, List, Union  # noqa: F401

from databuilder.models.neo4j_csv_serde import Neo4jCsvSerializable, NODE_KEY, \
    NODE_LABEL


class TestTableModel(Neo4jCsvSerializable):
    # type: (...) -> None
    """
    Hive table watermark result model.
    Each instance represents one row of hive watermark result.
    """
    LABEL = 'Table'
    KEY_FORMAT = '{db}://{cluster}.{schema}/{ tbl}'

    def __init__(self,
                 database,  # type: str
                 cluster,  # type: str
                 schema_name,  # type: str
                 table_name,  # type: str
                 table_desc,  # type: str
                 ):
        # type: (...) -> None
        self.database = database
        self.cluster = cluster
        self.schema_name = schema_name
        self.table_name = table_name
        self.table_desc = table_desc

        # currently we don't consider nested partitions
        self._node_iter = iter(self.create_nodes())
        self._relation_iter = iter(self.create_relation())

    def create_next_node(self):
        # type: (...) -> Union[Dict[str, Any], None]
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self):
        # type: (...) -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def create_nodes(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of Neo4j node records
        :return:
        """
        results = []

        results.append({
            NODE_KEY: '{db}://{cluster}.{schema}/{tbl}'.format(db=self.database,
                                                               cluster=self.cluster,
                                                               schema=self.schema_name,
                                                               tbl=self.table_name),
            NODE_LABEL: TestTableModel.LABEL,
            'table_desc': self.table_desc,
            'tbl_key': '{db}://{cluster}.{schema}/{tbl}'.format(db=self.database,
                                                                cluster=self.cluster,
                                                                schema=self.schema_name,
                                                                tbl=self.table_name)
        })
        return results

    def create_relation(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of relation map between watermark record with original hive table
        :return:
        """
        return []



from abc import ABCMeta, abstractmethod


class ElasticsearchDocument:
    """
    Base class for ElasticsearchDocument
    Each different resource ESDoc will be a subclass
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def to_json(self):
        # type: () -> str
        """
        Convert object to json for elasticsearch bulk upload
        Bulk load JSON format is defined here:
        https://www.elastic.co/guide/en/elasticsearch/reference/6.2/docs-bulk.html
        :return:
        """
        pass



from pyhocon import ConfigTree  # noqa: F401
from typing import Optional  # noqa: F401

from databuilder.transformer.base_transformer import Transformer
from databuilder.models.table_elasticsearch_document import TableESDocument
from databuilder.models.user_elasticsearch_document import UserESDocument
from databuilder.models.neo4j_data import Neo4jDataResult


class ElasticsearchDocumentTransformer(Transformer):
    """
    Transformer to convert Neo4j Graph data to Elasticsearch index document format
    """
    ELASTICSEARCH_INDEX_CONFIG_KEY = 'index'
    ELASTICSEARCH_DOC_CONFIG_KEY = 'doc_type'
    ELASTICSEARCH_RESOURCE_CONFIG_KEY = 'resource_type'
    RESOURCE_TYPE_MAPPING = {
        'table',
        'user'
    }

    def init(self, conf):
        # type: (ConfigTree) -> None
        self.conf = conf

        self.elasticsearch_index = self.conf.get_string(ElasticsearchDocumentTransformer.ELASTICSEARCH_INDEX_CONFIG_KEY)
        self.elasticsearch_type = self.conf.get_string(ElasticsearchDocumentTransformer.ELASTICSEARCH_DOC_CONFIG_KEY)
        # values: table, user
        self.elasticsearch_resource_type = self.conf.get_string(ElasticsearchDocumentTransformer.
                                                                ELASTICSEARCH_RESOURCE_CONFIG_KEY,
                                                                'table')

    def transform(self, record):
        # type: (Neo4jDataResult) -> Optional[ElasticsearchDocument]
        if not record:
            return None

        if not isinstance(record, Neo4jDataResult):
            raise Exception("ElasticsearchDocumentTransformer expects record of type 'Neo4jDataResult'!")

        if self.elasticsearch_resource_type.lower() not in \
                ElasticsearchDocumentTransformer.RESOURCE_TYPE_MAPPING:
            raise Exception('resource type needs to define in RESOURCE_TYPE_MAPPING')

        if self.elasticsearch_resource_type == 'table':
            elasticsearch_obj = TableESDocument(elasticsearch_index=self.elasticsearch_index,
                                                elasticsearch_type=self.elasticsearch_type,
                                                database=record.database,
                                                cluster=record.cluster,
                                                schema_name=record.schema_name,
                                                table_name=record.table_name,
                                                table_key=record.table_key,
                                                table_description=record.table_description,
                                                table_last_updated_epoch=record.table_last_updated_epoch,
                                                column_names=record.column_names,
                                                column_descriptions=record.column_descriptions,
                                                total_usage=record.total_usage,
                                                unique_usage=record.unique_usage,
                                                tag_names=record.tag_names)
        elif self.elasticsearch_resource_type == 'user':
            elasticsearch_obj = UserESDocument(elasticsearch_index=self.elasticsearch_index,
                                               elasticsearch_type=self.elasticsearch_type,
                                               email=record.email,
                                               first_name=record.first_name,
                                               last_name=record.last_name,
                                               name=record.name,
                                               github_username=record.github_username,
                                               team_name=record.team_name,
                                               employee_type=record.employee_type,
                                               manager_email=record.manager_email,
                                               slack_id=record.slack_id,
                                               is_active=record.is_active,
                                               total_read=record.total_read,
                                               total_own=record.total_own,
                                               total_follow=record.total_follow,
                                               )
        else:
            raise NotImplementedError()

        return elasticsearch_obj

    def get_scope(self):
        # type: () -> str
        return 'transformer.elasticsearch'



from typing import Any, Dict, List, Union  # noqa: F401

from databuilder.models.neo4j_csv_serde import Neo4jCsvSerializable, NODE_KEY, \
    NODE_LABEL, RELATION_START_KEY, RELATION_START_LABEL, RELATION_END_KEY, \
    RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE

from databuilder.models.user import User


class TableOwner(Neo4jCsvSerializable):
    # type: (...) -> None
    """
    Hive table owner model.
    """
    OWNER_TABLE_RELATION_TYPE = 'OWNER_OF'
    TABLE_OWNER_RELATION_TYPE = 'OWNER'

    def __init__(self,
                 db_name,  # type: str
                 schema_name,  # type: str
                 table_name,  # type: str
                 owners,  # type: List
                 cluster='gold',  # type: str
                 ):
        # type: (...) -> None
        self.db = db_name.lower()
        self.schema = schema_name.lower()
        self.table = table_name.lower()
        self.owners = owners

        self.cluster = cluster.lower()
        self._node_iter = iter(self.create_nodes())
        self._relation_iter = iter(self.create_relation())

    def create_next_node(self):
        # type: (...) -> Union[Dict[str, Any], None]
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self):
        # type: (...) -> Union[Dict[str, Any], None]
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def get_owner_model_key(self, owner  # type: str
                            ):
        # type: (...) -> str
        return User.USER_NODE_KEY_FORMAT.format(email=owner)

    def get_metadata_model_key(self):
        # type: (...) -> str
        return '{db}://{cluster}.{schema}/{table}'.format(db=self.db,
                                                          cluster=self.cluster,
                                                          schema=self.schema,
                                                          table=self.table)

    def create_nodes(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of Neo4j node records
        :return:
        """
        results = []
        for owner in self.owners:
            if owner:
                results.append({
                    NODE_KEY: self.get_owner_model_key(owner),
                    NODE_LABEL: User.USER_NODE_LABEL,
                    User.USER_NODE_EMAIL: owner
                })
        return results

    def create_relation(self):
        # type: () -> List[Dict[str, Any]]
        """
        Create a list of relation map between owner record with original hive table
        :return:
        """
        results = []
        for owner in self.owners:
            results.append({
                RELATION_START_KEY: self.get_owner_model_key(owner),
                RELATION_START_LABEL: User.USER_NODE_LABEL,
                RELATION_END_KEY: self.get_metadata_model_key(),
                RELATION_END_LABEL: 'Table',
                RELATION_TYPE: TableOwner.OWNER_TABLE_RELATION_TYPE,
                RELATION_REVERSE_TYPE: TableOwner.TABLE_OWNER_RELATION_TYPE
            })

        return results

    def __repr__(self):
        # type: () -> str
        return 'TableOwner({!r}, {!r}, {!r}, {!r}, {!r})'.format(self.db,
                                                                 self.cluster,
                                                                 self.schema,
                                                                 self.table,
                                                                 self.owners)



import abc

import six
from typing import Dict, Set, Any, Union  # noqa: F401

NODE_KEY = 'KEY'
NODE_LABEL = 'LABEL'
NODE_REQUIRED_HEADERS = {NODE_LABEL, NODE_KEY}

RELATION_START_KEY = 'START_KEY'
RELATION_START_LABEL = 'START_LABEL'
RELATION_END_KEY = 'END_KEY'
RELATION_END_LABEL = 'END_LABEL'
RELATION_TYPE = 'TYPE'
RELATION_REVERSE_TYPE = 'REVERSE_TYPE'
RELATION_REQUIRED_HEADERS = {RELATION_START_KEY, RELATION_START_LABEL,
                             RELATION_END_KEY, RELATION_END_LABEL,
                             RELATION_TYPE, RELATION_REVERSE_TYPE}

LABELS = {NODE_LABEL, RELATION_START_LABEL, RELATION_END_LABEL}
TYPES = {RELATION_TYPE, RELATION_REVERSE_TYPE}


@six.add_metaclass(abc.ABCMeta)
class Neo4jCsvSerializable(object):
    """
    A Serializable abstract class asks subclass to implement next node or
    next relation in dict form so that it can be serialized to CSV file.

    Any model class that needs to be pushed to Neo4j should inherit this class.
    """
    def __init__(self):
        # type: () -> None
        pass

    @abc.abstractmethod
    def create_next_node(self):
        # type: () -> Union[Dict[str, Any], None]
        """
        Creates dict where keys represent header in CSV and value represents
        row in CSV file. Should the class could have different types of
        nodes that it needs to serialize, it just needs to provide dict with
        different header -- the one who consumes this class figures it out and
        serialize to different file.

        Node is Neo4j's term of Vertex in Graph. More information on
        https://neo4j.com/docs/developer-manual/current/introduction/
        graphdb-concepts/
        :return: a dict or None if no more record to serialize
        """
        raise NotImplementedError

    @abc.abstractmethod
    def create_next_relation(self):
        # type: () -> Union[Dict[str, Any], None]
        """
        Creates dict where keys represent header in CSV and value represents
        row in CSV file. Should the class could have different types of
        relations that it needs to serialize, it just needs to provide dict
        with different header -- the one who consumes this class figures it
        out and serialize to different file.

        Relationship is Neo4j's term of Edge in Graph. More information on
        https://neo4j.com/docs/developer-manual/current/introduction/
        graphdb-concepts/
        :return: a dict or None if no more record to serialize
        """
        raise NotImplementedError

    def next_node(self):
        # type: () -> Union[Dict[str, Any], None]
        """
        Provides node(vertex) in dict form.
        Note that subsequent call can create different header (dict.keys())
        which implicitly mean that it needs to be serialized in different
        CSV file (as CSV is in fixed header)
        :return: Non-nested dict where key is CSV header and each value
        is a column
        """
        node_dict = self.create_next_node()
        if not node_dict:
            return None

        self._validate(NODE_REQUIRED_HEADERS, node_dict)
        return node_dict

    def next_relation(self):
        # type: () -> Union[Dict[str, Any], None]
        """
        Provides relation(edge) in dict form.
        Note that subsequent call can create different header (dict.keys())
        which implicitly mean that it needs to be serialized in different
        CSV file (as CSV is in fixed header)
        :return: Non-nested dict where key is CSV header and each value
        is a column
        """
        relation_dict = self.create_next_relation()
        if not relation_dict:
            return None

        self._validate(RELATION_REQUIRED_HEADERS, relation_dict)
        return relation_dict

    def _validate(self, required_set, val_dict):
        # type: (Set[str], Dict[str, Any]) -> None
        """
        Validates dict that represents CSV header and a row.
         - Checks if it has required headers for either Node or Relation
         - Checks value of LABEL if only first character is upper case
         - Checks value of TYPE if it's all upper case characters

        :param required_set:
        :param val_dict:
        :return:
        """
        required_count = 0
        for header_col, val_col in \
                ((header_col, val_col) for header_col, val_col
                 in six.iteritems(val_dict) if header_col in required_set):
            required_count += 1

            if header_col in LABELS:
                if not val_col.istitle():
                    raise RuntimeError(
                        'LABEL should only have upper case character on its '
                        'first one: {}'.format(val_col))
            elif header_col in TYPES:
                if not val_col == val_col.upper():
                    raise RuntimeError(
                        'TYPE needs to be upper case: '.format(val_col))

        if required_count != len(required_set):
            raise RuntimeError(
                'Required header missing. Required: {} , Header: {}'.format(
                    required_set, val_dict.keys()))



import textwrap
from typing import Any  # noqa: F401

from pyhocon import ConfigTree  # noqa: F401

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.neo4j_extractor import Neo4jExtractor
from databuilder.publisher.neo4j_csv_publisher import JOB_PUBLISH_TAG


class Neo4jSearchDataExtractor(Extractor):
    """
    Extractor to fetch data required to support search from Neo4j graph database
    Use Neo4jExtractor extractor class
    """
    CYPHER_QUERY_CONFIG_KEY = 'cypher_query'

    DEFAULT_NEO4J_CYPHER_QUERY = textwrap.dedent(
        """
        MATCH (db:Database)<-[:CLUSTER_OF]-(cluster:Cluster)<-[:SCHEMA_OF]-(schema:Schema)<-[:TABLE_OF]-(table:Table)
        {publish_tag_filter}
        OPTIONAL MATCH (table)-[:DESCRIPTION]->(table_description:Description)
        OPTIONAL MATCH (table)-[read:READ_BY]->(user:User)
        OPTIONAL MATCH (table)-[:COLUMN]->(cols:Column)
        OPTIONAL MATCH (cols)-[:DESCRIPTION]->(col_description:Description)
        OPTIONAL MATCH (table)-[:TAGGED_BY]->(tags:Tag)
        OPTIONAL MATCH (table)-[:LAST_UPDATED_AT]->(time_stamp:Timestamp)
        RETURN db.name as database, cluster.name AS cluster, schema.name AS schema_name,
        table.name AS table_name, table.key AS table_key, table_description.description AS table_description,
        time_stamp.last_updated_timestamp AS table_last_updated_epoch,
        EXTRACT(c in COLLECT(DISTINCT cols)| c.name) AS column_names,
        EXTRACT(cd IN COLLECT(DISTINCT col_description)| cd.description) AS column_descriptions,
        REDUCE(sum_r = 0, r in COLLECT(DISTINCT read)| sum_r + r.read_count) AS total_usage,
        COUNT(DISTINCT user.email) as unique_usage,
        COLLECT(DISTINCT tags.key) as tag_names
        ORDER BY table.name;
        """
    )

    def init(self, conf):
        # type: (ConfigTree) -> None
        """
        Initialize Neo4jExtractor object from configuration and use that for extraction
        """
        self.conf = conf

        # extract cypher query from conf, if specified, else use default query
        if Neo4jSearchDataExtractor.CYPHER_QUERY_CONFIG_KEY in conf:
            self.cypher_query = conf.get_string(Neo4jSearchDataExtractor.CYPHER_QUERY_CONFIG_KEY)
        else:
            self.cypher_query = self._add_publish_tag_filter(conf.get_string(JOB_PUBLISH_TAG, ''),
                                                             Neo4jSearchDataExtractor.DEFAULT_NEO4J_CYPHER_QUERY)

        self.neo4j_extractor = Neo4jExtractor()
        # write the cypher query in configs in Neo4jExtractor scope
        key = self.neo4j_extractor.get_scope() + '.' + Neo4jExtractor.CYPHER_QUERY_CONFIG_KEY
        self.conf.put(key, self.cypher_query)
        # initialize neo4j_extractor from configs
        self.neo4j_extractor.init(Scoped.get_scoped_conf(self.conf, self.neo4j_extractor.get_scope()))

    def close(self):
        # type: () -> None
        """
        Use close() method specified by neo4j_extractor
        to close connection to neo4j cluster
        """
        self.neo4j_extractor.close()

    def extract(self):
        # type: () -> Any
        """
        Invoke extract() method defined by neo4j_extractor
        """
        return self.neo4j_extractor.extract()

    def get_scope(self):
        # type: () -> str
        return 'extractor.search_data'

    def _add_publish_tag_filter(self, publish_tag, cypher_query):
        """
        Adds publish tag filter into Cypher query
        :param publish_tag: value of publish tag.
        :param cypher_query:
        :return:
        """
        # type: (str, str) -> str
        if not publish_tag:
            publish_tag_filter = ''
        else:
            publish_tag_filter = """WHERE table.published_tag = '{}'""".format(publish_tag)

        return cypher_query.format(publish_tag_filter=publish_tag_filter)



import json
from typing import List, Optional  # noqa: F401

from databuilder.models.elasticsearch_document import ElasticsearchDocument


class UserESDocument(ElasticsearchDocument):
    """
    Schema for the Search index document for user
    """
    def __init__(self,
                 elasticsearch_index,  # type: str
                 elasticsearch_type,   # type: str
                 email,  # type: str
                 first_name,  # type: str
                 last_name,  # type: str
                 name,  # type: str
                 github_username,  # type: str
                 team_name,  # type: str
                 employee_type,  # type: str
                 manager_email,  # type: str
                 slack_id,  # type: str
                 is_active,  # type: bool
                 total_read,  # type: int
                 total_own,  # type: int
                 total_follow,  # type: int
                 ):
        # type: (...) -> None
        self.elasticsearch_index = elasticsearch_index
        self.elasticsearch_type = elasticsearch_type
        self.email = email
        self.first_name = first_name
        self.last_name = last_name
        self.name = name
        self.github_username = github_username
        self.team_name = team_name
        self.employee_type = employee_type
        self.manager_email = manager_email
        self.slack_id = slack_id
        self.is_active = is_active
        self.total_read = total_read
        self.total_own = total_own
        self.total_follow = total_follow

    def to_json(self):
        # type: () -> str
        """
        Convert object to json for elasticsearch bulk upload
        Bulk load JSON format is defined here:
        https://www.elastic.co/guide/en/elasticsearch/reference/6.2/docs-bulk.html
        :return:
        """
        index_row = dict(index=dict(_index=self.elasticsearch_index,
                                    _type=self.elasticsearch_type))
        data = json.dumps(index_row) + "\n"

        # convert rest of the object
        obj_dict = {k: v for k, v in sorted(self.__dict__.items())
                    if k not in ['elasticsearch_index', 'elasticsearch_type']}
        data += json.dumps(obj_dict) + "\n"

        return data



import json
from typing import List, Optional  # noqa: F401

from databuilder.models.elasticsearch_document import ElasticsearchDocument


class TableESDocument(ElasticsearchDocument):
    """
    Schema for the Search index document
    """
    def __init__(self,
                 elasticsearch_index,  # type: str
                 elasticsearch_type,   # type: str
                 database,  # type: str
                 cluster,  # type: str
                 schema_name,  # type: str
                 table_name,  # type: str
                 table_key,  # type: str
                 table_description,  # type: str
                 table_last_updated_epoch,  # type: Optional[int]
                 column_names,  # type: List[str]
                 column_descriptions,  # type: List[str]
                 total_usage,  # type: int
                 unique_usage,  # type: int
                 tag_names,  # type: List[str]
                 ):
        # type: (...) -> None
        self.elasticsearch_index = elasticsearch_index
        self.elasticsearch_type = elasticsearch_type
        self.database = database
        self.cluster = cluster
        self.schema_name = schema_name
        self.table_name = table_name
        self.table_key = table_key
        self.table_description = table_description
        self.table_last_updated_epoch = table_last_updated_epoch
        self.column_names = column_names
        self.column_descriptions = column_descriptions
        self.total_usage = total_usage
        self.unique_usage = unique_usage
        # todo: will include tag_type once we have better understanding from UI flow.
        self.tag_names = tag_names

    def to_json(self):
        # type: () -> str
        """
        Convert object to json for elasticsearch bulk upload
        Bulk load JSON format is defined here:
        https://www.elastic.co/guide/en/elasticsearch/reference/6.2/docs-bulk.html
        :return:
        """
        index_row = dict(index=dict(_index=self.elasticsearch_index,
                                    _type=self.elasticsearch_type))
        data = json.dumps(index_row) + "\n"

        # convert rest of the object
        obj_dict = {k: v for k, v in sorted(self.__dict__.items())
                    if k not in ['elasticsearch_index', 'elasticsearch_type']}
        data += json.dumps(obj_dict) + "\n"

        return data



from setuptools import setup, find_packages


__version__ = '1.2.1'


setup(
    name='amundsen-databuilder',
    version=__version__,
    description='Amundsen Data builder',
    url='https://www.github.com/lyft/amundsendatabuilder',
    maintainer='Lyft',
    maintainer_email='dev@lyft.com',
    packages=find_packages(exclude=['tests*']),
    dependency_links=[],
    install_requires=[
        # Packages in here should rarely be pinned. This is because these
        # packages (at the specified version) are required for project
        # consuming this library. By pinning to a specific version you are the
        # number of projects that can consume this or forcing them to
        # upgrade/downgrade any dependencies pinned here in their project.
        #
        # Generally packages listed here are pinned to a major version range.
        #
        # e.g.
        # Python FooBar package for foobaring
        # pyfoobar>=1.0, <2.0
        #
        # This will allow for any consuming projects to use this library as
        # long as they have a version of pyfoobar equal to or greater than 1.x
        # and less than 2.x installed.
        'pyhocon==0.3.42',
        'pytest==3.6.0',
        'six',
        'neo4j-driver==1.7.2',
        'antlr4-python2-runtime==4.7.1',
        'statsd==3.2.1',
        'retrying==1.3.3'
    ],
    extras_require={
        ':python_version=="2.7"': ['typing>=3.6'],  # allow typehinting PY2
    },
)
